-------------------------------------------------------------------------------
--! SPDX-FileCopyrightText: 2022 CERN (home.cern)
--
--! SPDX-License-Identifier: CERN-OHL-W-2.0+
-------------------------------------------------------------------------------
--! Title      : mt_profip_translator
--! Project    : ProFIP
-------------------------------------------------------------------------------
--! File       : mt_profip_translator.vhd
--! Author     : Konstantinos Blantos <konstantinos.blantos@cern.ch>
--! Company    : CERN (BE-CEM-EDL)
--! Created    : 22-02-2023
--! Last update:
--! Platform   : FPGA-generic
--! Standard   : VHDL'08
-------------------------------------------------------------------------------
-- Description: Responsible of received the MOSI data from ERTEC through SPI
--              and transmitting MISO data from Mockturtle. For communication
--              with MockTurtle, Remote Message Queues are used.
-------------------------------------------------------------------------------

--=============================================================================
--!                            Libraries & Packages                           --
--=============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mock_turtle_pkg.all;
use work.gencores_pkg.all;
use work.mt_mqueue_pkg.all;

--=============================================================================
--!               Entity declaration for mt_profip_translator                --
--=============================================================================

entity mt_profip_translator is
  generic (
    -- Data size and SPI mode
    g_data_width     : natural;
    g_cpol           : natural range 0 to 1;
    g_cpha           : natural range 0 to 1;
    -- Crossing Clock Domain enable for 1
    g_cdc_enable     : natural range 0 to 1);
  port (
    clk_i            : in  std_logic; -- FPGA clock
    rst_n_i          : in  std_logic; -- FPGA reset
    --  rmq_status_o      : out std_logic_vector(6 downto 0);
    rmq_id_o         : out std_logic_vector(7 downto 0); -- contains the core and slot of RMQ

    -- RX (MSTRFIP TRANSLATOR -> RMQ)
    rmq_src_i        : in  t_mt_stream_source_in;  -- (ready, pkt_ready)
    rmq_src_o        : out t_mt_stream_source_out; -- (data, hdr, valid, last, error)
    rmq_src_config_i : in  t_mt_stream_config_out; -- (adr, dat, we)
    rmq_src_config_o : out t_mt_stream_config_in;  -- (dat)

    -- TX (RMQ -> MSTRFIP_TRANSLATOR)
    rmq_snk_i        : in  t_mt_stream_sink_in_array2d; -- (data, hdr, valid, last, error)
    rmq_snk_o        : out t_mt_stream_sink_out_array2d;   -- (ready, pkt_ready)
    rmq_snk_config_i : in  t_mt_stream_config_out; -- (adr, dat, we)
    rmq_snk_config_o : out t_mt_stream_config_in;  -- (dat)

    -- SPI interface
    spi_clk_i        : in  std_logic; -- SPI clock
    spi_cs_n_i       : in  std_logic; -- SPI chip select, active LOW
    spi_mosi_i       : in  std_logic; -- SPI serial data, master out slave in
    spi_miso_o       : out std_logic); -- SPI serial data, master in slave out
end entity mt_profip_translator;

--==============================================================================
--!                           Architecture declaration                        --
--==============================================================================

architecture rtl of mt_profip_translator is

  --! FSM types
  type t_wr_state is (IDLE, BASIC_INFO, HEADER_1, HEADER_2, WAIT_HEADER,
    DATA_1, DATA_2, WAIT_DATA, CHECKSUM);

  type t_rd_state is (IDLE, WAIT_HEADER, RD_HEADER, PRE_LOAD, LOAD_HWORD, LOAD_LWORD, DATA, ERROR_STATE, END_OF_FRAME);

  --! FIFO array
  type t_miso_fifo_data_array is array (6 downto 0) of std_logic_vector(g_data_width+3 downto 0);
  type t_miso_fifo_flag_array is array (6 downto 0) of std_logic;

  signal s_wr_state : t_wr_state;
  signal s_rd_state : t_rd_state;

  --! Internal signals
  signal spi_mosi_data      : std_logic_vector(g_data_width - 1 downto 0) := (others => '0');
  signal spi_miso_data      : std_logic_vector(g_data_width - 1 downto 0) := (others => '0');
  signal mosi_data          : std_logic_vector(g_data_width - 1 downto 0) := (others => '0');
  signal wr_control         : std_logic_vector(7 downto 0); -- stores the CONTROL byte from MOSI
  signal wr_rmq_id          : std_logic_vector(7 downto 0); -- stores the RMQ_ID  byte from MOSI
  signal wr_data_len        : std_logic_vector(7 downto 0); -- stores the LENGTH  byte from MOSI
  signal wr_data_cnt        : unsigned(7 downto 0); -- counts the number of data bytes from MOSI
  signal wr_header_cnt      : unsigned(1 downto 0); -- counts the number of header bytes from MOSI
  signal wr_header          : std_logic_vector(15 downto 0);
  signal wr_data            : std_logic_vector(15 downto 0);
  signal spi_data_valid_i   : std_logic;
  signal spi_data_valid_o   : std_logic;
  signal s_dv_o             : std_logic;
  signal s_ready            : std_logic;
  signal s_operation        : std_logic;
  signal s_oper_en          : std_logic;
  signal valid              : std_logic;
  signal fail               : std_logic;
  signal rd_en              : std_logic;
  signal spi_busy           : std_logic;
  signal spi_ready          : std_logic;
  signal miso_data_reg      : std_logic_vector(31 downto 0);
  signal miso_last          : std_logic;
  signal s_data_valid       : std_logic;
  signal s_rmq_id           : natural range 0 to 6;

  --! MOSI async FIFO signals
  signal mosi_fifo_data     : std_logic_vector(g_data_width - 1 downto 0) := (others => '0');
  signal mosi_wr_en         : std_logic;
  signal mosi_rd_en         : std_logic;
  signal mosi_fifo_wr_empty : std_logic;
  signal mosi_fifo_wr_full  : std_logic;
  signal mosi_fifo_rd_empty : std_logic;
  signal mosi_fifo_rd_full  : std_logic;
  --! MISO async FIFO signals
  signal miso_fifo_data_in  : t_miso_fifo_data_array;
  signal miso_fifo_data_out : t_miso_fifo_data_array; 
  signal miso_wr_en         : t_miso_fifo_flag_array;
  signal miso_rd_en         : t_miso_fifo_flag_array;
  signal miso_fifo_wr_empty : t_miso_fifo_flag_array;
  signal miso_fifo_wr_full  : t_miso_fifo_flag_array;
  signal miso_fifo_rd_empty : t_miso_fifo_flag_array;
  signal miso_fifo_rd_full  : t_miso_fifo_flag_array;

begin

  ----------------------------------------------------------------------------
  --! Included components in MasterFIP Translator
  ----------------------------------------------------------------------------

  --! SPI Slave component
  cmp_spi_slave : entity work.spi_slave
    generic map(
      g_cpol       => g_cpol,
      g_cpha       => g_cpha,
      g_data_width => g_data_width)
    port map(
      clk_i        => clk_i,
      rst_n_i      => rst_n_i,
      spi_clk_i    => spi_clk_i,
      spi_cs_n_i   => spi_cs_n_i,
      spi_mosi_i   => spi_mosi_i,
      spi_miso_o   => spi_miso_o,
      data_i       => spi_miso_data,
      data_valid_i => spi_data_valid_i,
      data_o       => spi_mosi_data,
      data_valid_o => spi_data_valid_o,
      ready_o      => spi_ready,
      busy_o       => spi_busy,
      operation_o  => s_operation
    );

  gen_sync_fifo : if g_cdc_enable = 0 generate
  begin

    --! In case there is no CDC enable, use the synchronous FIFO
    cmp_mosi_sync_fifo : entity work.inferred_sync_fifo
    generic map (
        g_data_width             => g_data_width,
        g_size                   => 4,
        g_show_ahead             => true,
        g_show_ahead_legacy_mode => true,
        g_with_empty             => true,
        g_with_full              => true,
        g_with_almost_empty      => false,
        g_with_almost_full       => false,
        g_with_count             => false,
        g_almost_empty_threshold => 0,
        g_almost_full_threshold  => 0,
        g_register_flag_outputs  => true)
    port map (
        rst_n_i                  => rst_n_i,
        clk_i                    => clk_i,
        d_i                      => spi_mosi_data,
        we_i                     => mosi_wr_en,
        q_o                      => mosi_fifo_data,
        rd_i                     => mosi_rd_en,
        empty_o                  => mosi_fifo_rd_empty,
        full_o                   => mosi_fifo_wr_full,
        almost_empty_o           => open,
        almost_full_o            => open,
        count_o                  => open);


    --! Multiple Sync FIFOs are used because there are multiple inputs
    gen_sync_miso_fifo_per_cpu_per_slot : for i in 0 to 6 generate
    begin

      --! MISO synchronous FIFO
      cmp_miso_sync_fifo : entity work.inferred_sync_fifo
      generic map (
          g_data_width             => g_data_width + 4,
          g_size                   => 128,
          g_show_ahead             => true,
          g_show_ahead_legacy_mode => true,
          g_with_empty             => true,
          g_with_full              => true,
          g_with_almost_empty      => false,
          g_with_almost_full       => false,
          g_with_count             => false,
          g_almost_empty_threshold => 0,
          g_almost_full_threshold  => 0,
          g_register_flag_outputs  => true)
      port map (
          rst_n_i                  => rst_n_i,
          clk_i                    => clk_i,
          d_i                      => miso_fifo_data_in(i),
          we_i                     => miso_wr_en(i),
          q_o                      => miso_fifo_data_out(i),
          rd_i                     => miso_rd_en(i),
          empty_o                  => miso_fifo_rd_empty(i),
          full_o                   => miso_fifo_wr_full(i),
          almost_empty_o           => open,
          almost_full_o            => open,
          count_o                  => open);

    end generate gen_sync_miso_fifo_per_cpu_per_slot;

  end generate gen_sync_fifo;

  gen_async_fifo : if g_cdc_enable = 1 generate
  begin
      --! Asynchronous FIFO used for fixing the CDC between
      --! SPI Slave and Translator, because there are both
      --! working under different clock domains.
      cmp_mosi_async_fifo : entity work.inferred_async_fifo
        generic map(
          g_data_width             => g_data_width,
          g_size                   => 4,
          g_show_ahead             => TRUE,
          g_with_rd_empty          => TRUE,
          g_with_rd_full           => FALSE,
          g_with_rd_almost_empty   => FALSE,
          g_with_rd_almost_full    => FALSE,
          g_with_rd_count          => FALSE,
          g_with_wr_empty          => FALSE,
          g_with_wr_full           => TRUE,
          g_with_wr_almost_empty   => FALSE,
          g_with_wr_almost_full    => FALSE,
          g_with_wr_count          => FALSE,
          g_almost_empty_threshold => 0,
          g_almost_full_threshold  => 1023)
        port map(
          rst_n_i                  => rst_n_i,
          clk_wr_i                 => clk_i,
          d_i                      => spi_mosi_data,
          we_i                     => mosi_wr_en,
          wr_empty_o               => mosi_fifo_wr_empty,
          wr_full_o                => mosi_fifo_wr_full,
          wr_almost_empty_o        => open,
          wr_almost_full_o         => open,
          wr_count_o               => open,
          clk_rd_i                 => clk_i,
          q_o                      => mosi_fifo_data,
          rd_i                     => mosi_rd_en,
          rd_empty_o               => mosi_fifo_rd_empty,
          rd_full_o                => mosi_fifo_rd_full,
          rd_almost_empty_o        => open,
          rd_almost_full_o         => open,
          rd_count_o               => open);

      --! Asynchronous FIFO used for fixing the CDC between
      --! Translator and SPI Slave, because there are both
      --! working under different clock domains. Data width
      --! of this FIFO will be different because now data
      --! contains the control signals (hdr,valid,error,last)
          --! Multiple Sync FIFOs are used because there are multiple inputs
    gen_async_miso_fifo_per_cpu_per_slot : for i in 0 to 6 generate
    begin
      cmp_miso_async_fifo : entity work.inferred_async_fifo
        generic map(
          g_data_width             => g_data_width + 4,
          g_size                   => 4,
          g_show_ahead             => TRUE,
          g_with_rd_empty          => TRUE,
          g_with_rd_full           => FALSE,
          g_with_rd_almost_empty   => FALSE,
          g_with_rd_almost_full    => FALSE,
          g_with_rd_count          => FALSE,
          g_with_wr_empty          => FALSE,
          g_with_wr_full           => TRUE,
          g_with_wr_almost_empty   => FALSE,
          g_with_wr_almost_full    => FALSE,
          g_with_wr_count          => FALSE,
          g_almost_empty_threshold => 0,
          g_almost_full_threshold  => 1023)
        port map(
          rst_n_i                  => rst_n_i,
          clk_wr_i                 => clk_i,
          d_i                      => miso_fifo_data_in(i),
          we_i                     => miso_wr_en(i),
          wr_empty_o               => miso_fifo_wr_empty(i),
          wr_full_o                => miso_fifo_wr_full(i),
          wr_almost_empty_o        => open,
          wr_almost_full_o         => open,
          wr_count_o               => open,
          clk_rd_i                 => clk_i,
          q_o                      => miso_fifo_data_out(i),
          rd_i                     => miso_rd_en(i),
          rd_empty_o               => miso_fifo_rd_empty(i),
          rd_full_o                => miso_fifo_rd_full(i),
          rd_almost_empty_o        => open,
          rd_almost_full_o         => open,
          rd_count_o               => open);
    end generate gen_async_miso_fifo_per_cpu_per_slot;
  
  end generate gen_async_fifo;

  --! t_mt_stream_config_in and out will not be used
  rmq_src_config_o <= c_MT_DUMMY_EP_CONFIG_IN;
  rmq_snk_config_o <= c_MT_DUMMY_EP_CONFIG_IN;

  --! Final and synchronized MOSI data that will go
  --! to RMQ is the data taken from SPI Slave when
  --! the read enable of MOSI FIFO is HIGH
  p_mosi_data : process (clk_i, rst_n_i)
  begin
      if rst_n_i = '0' then
          mosi_data <= (others => '0');
      elsif rising_edge(clk_i) then
          if mosi_rd_en = '1' then
              mosi_data <= mosi_fifo_data;
          else
              mosi_data <= mosi_data;
          end if;
      end if;
  end process;

  --! The format of the data is the following:
  --! 35   - hdr
  --! 34   - valid
  --! 33   - last
  --! 32   - error
  --! 31:0 - data

  --! For Core 0, slots 0 to 5 assignment
--  process (rmq_snk_i)
--  begin
--    miso_fifo_data_assignment : for j in 0 to 5 loop
--      miso_fifo_data_in(j) <= (rmq_snk_i(0)(j).hdr & rmq_snk_i(0)(j).valid & rmq_snk_i(0)(j).last & rmq_snk_i(0)(j).error & rmq_snk_i(0)(j).data);
--    end loop;
--  end process;

  --! Assignment of the last FIFO for core 1 slot 0
  miso_fifo_data_in(0) <= (rmq_snk_i(0)(0).hdr & rmq_snk_i(0)(0).valid & rmq_snk_i(0)(0).last & rmq_snk_i(0)(0).error & rmq_snk_i(0)(0).data);
  miso_fifo_data_in(1) <= (rmq_snk_i(0)(1).hdr & rmq_snk_i(0)(1).valid & rmq_snk_i(0)(1).last & rmq_snk_i(0)(1).error & rmq_snk_i(0)(1).data);
  miso_fifo_data_in(2) <= (rmq_snk_i(0)(2).hdr & rmq_snk_i(0)(2).valid & rmq_snk_i(0)(2).last & rmq_snk_i(0)(2).error & rmq_snk_i(0)(2).data);
  miso_fifo_data_in(3) <= (rmq_snk_i(0)(3).hdr & rmq_snk_i(0)(3).valid & rmq_snk_i(0)(3).last & rmq_snk_i(0)(3).error & rmq_snk_i(0)(3).data);
  miso_fifo_data_in(4) <= (rmq_snk_i(0)(4).hdr & rmq_snk_i(0)(4).valid & rmq_snk_i(0)(4).last & rmq_snk_i(0)(4).error & rmq_snk_i(0)(4).data);
  miso_fifo_data_in(5) <= (rmq_snk_i(0)(5).hdr & rmq_snk_i(0)(5).valid & rmq_snk_i(0)(5).last & rmq_snk_i(0)(5).error & rmq_snk_i(0)(5).data);
  miso_fifo_data_in(6) <= (rmq_snk_i(1)(0).hdr & rmq_snk_i(1)(0).valid & rmq_snk_i(1)(0).last & rmq_snk_i(1)(0).error & rmq_snk_i(1)(0).data);


  --! Selection of WRITE or READ to/from RMQ: this is the first 32-bits that ERTEC sends when the CS goes LOW
  --! this is selected by wr_control(0). So we have:
  --!    wr_control(0) = '0' --> WRITE to RMQ
  --!    wr_control(0) = '1' --> READ from RMQ

  ----------------------------------------------------------------------------
  --! RX:  MASTERFIP TRANSLATOR --> MT (RMQ)
  ----------------------------------------------------------------------------

  --! FSM for writing from ERTEC to RMQ. Exiting the IDLE state when source
  --! input is ready and there is valid data coming from MOSI FIFO. In the BASIC
  --! INFO state the received data consist of 3 bytes (1 control byte, 1 RMQ_ID
  --! byte and 1 byte containg the Number of Bytes that the data contains). The
  --! In the HEADER state the header bytes assigned to src_o.data (3x32-bits).
  --! 4th byte is unused. Depending on the data len, received in the BASIC INFO,
  --! there is a counter which counts the number of packets of 32-bit data that
  --! received and is to be sent to RMQ. Once it is stops transmitting, last
  --! is asserted. CHECKSUM state, is the state that FSM will go for 1 clock cycle,
  --! and error will be asserted. Then, either it is an error or not, FSM will IDLE.
  --! NOTE: the incoming mosi data has 32-bit width, same as rmq data bus, but
  --! rmq discard half of them everytime, so that's why valid data should be send
  --! as a packet of 16-bits every time.
  p_wr_fsm : process (clk_i, rst_n_i, spi_cs_n_i)
  begin
    if (rst_n_i = '0') then
      s_wr_state <= IDLE;
      wr_control <= (others => '0');
      wr_rmq_id <= (others => '0');
      wr_data_len <= (others => '0');
      wr_header_cnt <= (others => '1');
      wr_data_cnt <= (others => '0');
      fail <= '0';
    elsif rising_edge(clk_i) then
      case s_wr_state is

          ----------------------------------------------------------------
          --! STATE IDLE
        when IDLE =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1') then
            s_wr_state    <= BASIC_INFO;
            wr_control    <= mosi_fifo_data(31 downto 24);
            wr_rmq_id     <= mosi_fifo_data(23 downto 16);
            wr_data_len   <= mosi_fifo_data(15 downto 8);
            wr_header_cnt <= (others => '1');
          else
            s_wr_state <= IDLE;
            wr_control <= (others => '0');
            wr_rmq_id <= (others => '0');
            wr_data_len <= (others => '0');
            wr_header_cnt <= (others => '1');
            wr_data_cnt <= (others => '0');
            fail <= '0';
          end if;

          ----------------------------------------------------------------
          --! STATE BASIC_INFO
        when BASIC_INFO =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1' and wr_control(0) = '0') then --KB
            wr_data_cnt <= unsigned(wr_data_len); -- assign the width of payload in bytes
            s_wr_state <= HEADER_1;
          elsif (spi_cs_n_i = '1' or wr_control(0) = '1') then --KB: when we supposed to read, go to IDLE
            s_wr_state <= CHECKSUM;
            fail <= '1';
          else
            s_wr_state <= BASIC_INFO;
            fail <= '0';
          end if;

          ----------------------------------------------------------------
          --! STATE HEADER_1
        when HEADER_1 =>
          if (rmq_src_i.ready = '1') then
            if wr_header_cnt /= 0 then
              s_wr_state <= HEADER_2;
              wr_header_cnt <= wr_header_cnt - 1;
            else
              s_wr_state <= IDLE; --otherwise we will be stuck for ever here
            end if;
          elsif (spi_cs_n_i = '1') then
            fail <= '1';
            s_wr_state <= CHECKSUM;
          else
            s_wr_state <= HEADER_1;
            fail <= '0';
          end if;

          ----------------------------------------------------------------
          --! STATE HEADER_2
        when HEADER_2 =>
          if (rmq_src_i.ready = '1') then
            s_wr_state <= WAIT_HEADER;
          elsif (spi_cs_n_i = '1') then
            fail <= '1';
            s_wr_state <= CHECKSUM;
          else
            fail <= '0';
            s_wr_state <= HEADER_2;
          end if;

          ----------------------------------------------------------------
          --! STATE WAIT_HEADER
        when WAIT_HEADER =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1') then
            if wr_header_cnt = 0 then
              s_wr_state <= DATA_1;
            else
              s_wr_state <= HEADER_1;
            end if;
          elsif (spi_cs_n_i = '1') then
            fail <= '1';
            s_wr_state <= CHECKSUM;
          else
            fail <= '0';
            s_wr_state <= WAIT_HEADER;
          end if;

          ----------------------------------------------------------------
          --! STATE DATA_1
        when DATA_1 =>
          if (rmq_src_i.ready = '1') then
            if wr_data_cnt /= 0 then
              s_wr_state <= DATA_2;
              wr_data_cnt <= wr_data_cnt - 1;
            else
              s_wr_state <= IDLE; --otherwise we will be stuck for ever here
            end if;
          elsif (spi_cs_n_i = '1') then
            fail <= '1';
            s_wr_state <= CHECKSUM;
          else
            s_wr_state <= DATA_1;
            fail <= '0';
          end if;

          ----------------------------------------------------------------
          --! STATE DATA_2
        when DATA_2 =>
          if (rmq_src_i.ready = '1') then
            s_wr_state <= WAIT_DATA;
          elsif (spi_cs_n_i = '1') then
            fail <= '1';
            s_wr_state <= CHECKSUM;
          else
            fail <= '0';
            s_wr_state <= DATA_2;
          end if;

          ----------------------------------------------------------------
          --! STATE WAIT_DATA
        when WAIT_DATA =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1') then
            if wr_data_cnt = 0 then
              s_wr_state <= CHECKSUM;
            else
              s_wr_state <= DATA_1;
            end if;

          elsif (wr_data_cnt = 0) then
            s_wr_state <= CHECKSUM;
          else
            fail <= '0';
            s_wr_state <= WAIT_DATA;
          end if;

          ----------------------------------------------------------------
          --! STATE CHECKSUM
        when CHECKSUM =>
          fail <= '0';
          if spi_cs_n_i = '1' then
            s_wr_state <= IDLE;
          else
            s_wr_state <= CHECKSUM;
          end if;

          ----------------------------------------------------------------
          --! STATE OTHERS
        when others =>
          fail <= '0';
          s_wr_state <= IDLE;

      end case;
    end if;
  end process p_wr_fsm;

  --! Vector that contains information about the core and slot of RMQ
  rmq_id_o <= wr_rmq_id;

  --! RMQ protocol, requires that we should devide the header and payload data into 2 parts
  wr_header <= mosi_data(31 downto 16) when s_wr_state = HEADER_1 else
               mosi_data(15 downto 0) when s_wr_state = HEADER_2 else (others => '0');

  wr_data <= mosi_data(31 downto 16) when s_wr_state = DATA_1 else
             mosi_data(15 downto 0) when s_wr_state = DATA_2 else (others => '0');

  rmq_src_o.data <= (x"0000" & wr_header) when (s_wr_state = HEADER_1 or s_wr_state = HEADER_2) else
                    (x"0000" & wr_data) when (s_wr_state = DATA_1 or s_wr_state = DATA_2) else (others => '0');

  rmq_src_o.hdr <= '1' when (s_wr_state = HEADER_1 or s_wr_state = HEADER_2) else '0';

  valid <= '1' when ((s_wr_state = HEADER_1 or s_wr_state = HEADER_2) or (s_wr_state = DATA_1 or s_wr_state = DATA_2)) else '0';
  rmq_src_o.valid <= valid or fail; -- in case to catch the error, we need valid HIGH (!)
  rmq_src_o.last <= '1' when (wr_data_cnt = 0 and valid = '1') else '0';

  --! Assert error signal when less header or payload data has been sent by ERTEC (and CS is HIGH)
  rmq_src_o.error <= fail;

  --! Write enable of the MOSI FIFO
  mosi_wr_en <= '1' when (s_dv_o = '1' and mosi_fifo_wr_full = '0') else '0'; -- we are using the output valid from SPI Slave (it is in the clk domain of MT)

  --! Read enable of the MOSI FIFO
  mosi_rd_en <= '1' when (mosi_fifo_rd_empty = '0' and rmq_src_i.ready = '1') else '0';

  --! Capture the positive edge of the spi_data_valid_o
  cmp_dv_o_edge_detect : entity work.gc_edge_detect
    generic map(
      g_ASYNC_RST  => false,
      g_PULSE_EDGE => "positive",
      g_CLOCK_EDGE => "positive")
    port map(
      clk_i        => clk_i,
      rst_n_i      => rst_n_i,
      data_i       => spi_data_valid_o,
      pulse_o      => s_dv_o);

  ----------------------------------------------------------------------------
  --! TX: MT (RMQ) --> MASTERFIP TRANSLATOR
  ----------------------------------------------------------------------------

  --! FSM for reading data from RMQ. The basic idea is that we read the data from
  --! the MISO FIFO that it is already stored. If it is a header data
  p_rd_fsm : process (clk_i, rst_n_i)
  begin
    if rst_n_i = '0' then
      s_rd_state    <= IDLE;
      miso_data_reg <= (others => '0');
      s_data_valid  <= '0';
      miso_last     <= '0';

    elsif rising_edge(clk_i) then
      case s_rd_state is

          ----------------------------------------------------------------
          --! STATE IDLE
        when IDLE =>
          miso_last     <= '0';
          s_data_valid  <= '0';
          miso_data_reg <= (others => '0');
          if s_oper_en = '1' then
            s_rd_state <= WAIT_HEADER;
          else
            s_rd_state <= IDLE;
          end if;

          ----------------------------------------------------------------
          --! STATE WAIT_HEADER
        when WAIT_HEADER =>
          if s_ready = '1' and miso_fifo_rd_empty(s_rmq_id) = '0' and miso_fifo_data_out(s_rmq_id)(35 downto 32) = "1100" then
            s_data_valid  <= '1';
            miso_data_reg <= miso_fifo_data_out(s_rmq_id)(31 downto 0);
            s_rd_state    <= RD_HEADER;
          elsif miso_fifo_rd_empty(s_rmq_id) = '1' then
            s_rd_state <= ERROR_STATE;
          else
            miso_data_reg <= miso_data_reg;
            s_data_valid  <= '0';
            s_rd_state    <= WAIT_HEADER;
          end if;

          ----------------------------------------------------------------
          --! STATE RD_HEADER
        when RD_HEADER =>
          s_data_valid <= '0';
          s_rd_state   <= PRE_LOAD;

          ----------------------------------------------------------------
          --! STATE PRE_LOAD
        when PRE_LOAD =>
          s_data_valid <= '0';
          if miso_fifo_rd_empty(s_rmq_id) = '0' then
            s_rd_state <= LOAD_HWORD;
          else
            s_rd_state <= PRE_LOAD;
          end if;

          ----------------------------------------------------------------
          --! STATE LOAD_HWORD
        when LOAD_HWORD =>
          s_data_valid <= '0';
          if miso_fifo_rd_empty(s_rmq_id) = '0' then
            if miso_fifo_data_out(s_rmq_id)(34) = '1' and miso_fifo_data_out(s_rmq_id)(32) = '0' then -- valid & !error
              miso_data_reg(31 downto 16) <= miso_fifo_data_out(s_rmq_id)(15 downto 0);
              s_rd_state <= LOAD_LWORD;
            else
              s_rd_state <= ERROR_STATE;
            end if;
          else
            s_rd_state <= LOAD_HWORD;
          end if;

          ----------------------------------------------------------------
          --! STATE LOAD_LWORD
        when LOAD_LWORD =>
          if miso_fifo_rd_empty(s_rmq_id) = '0' then
            if miso_fifo_data_out(s_rmq_id)(34) = '1' and miso_fifo_data_out(s_rmq_id)(32) = '0' then -- valid & !error
              miso_data_reg(15 downto 0) <= miso_fifo_data_out(s_rmq_id)(15 downto 0);
              if miso_fifo_data_out(s_rmq_id)(33) = '1' then -- if last
                miso_last  <= '1';
                s_rd_state <= END_OF_FRAME;
              end if;
              s_rd_state <= DATA;
            else
              s_rd_state <= ERROR_STATE;
            end if;
          else
            s_rd_state <= LOAD_LWORD;
          end if;

          ----------------------------------------------------------------
          --! STATE DATA
        when DATA =>
          if s_ready = '1' then
            s_data_valid <= '1';
            if miso_last = '1' then
              s_rd_state <= END_OF_FRAME;
            else
              s_rd_state <= LOAD_HWORD;
            end if;
          else
            s_rd_state <= DATA;
          end if;

          ----------------------------------------------------------------
          --! STATE ERROR_STATE
        when ERROR_STATE =>
          s_data_valid  <= '1';
          miso_last     <= '0';
          miso_data_reg <= x"DEADCAFE";
          if spi_cs_n_i = '1' then
            s_rd_state <= IDLE;
          else
            s_rd_state <= ERROR_STATE;
          end if;

          ----------------------------------------------------------------
          --! STATE END_OF_FRAME
        when END_OF_FRAME =>
          s_data_valid  <= '0';
          miso_last     <= '0';
          miso_data_reg <= (others => '0');
          if spi_cs_n_i = '1' then
            s_rd_state <= IDLE;
          else
            s_rd_state <= END_OF_FRAME;
          end if;

          ----------------------------------------------------------------
          --! STATE DEFAULT
        when others =>
          s_rd_state <= IDLE;

          ----------------------------------------------------------------
      end case;
    end if;
  end process;

  --! Ready and packet ready for the multiple RMQs
  process (miso_fifo_wr_full,miso_fifo_rd_empty,rd_en,miso_fifo_data_in)
  begin

      rmq_snk_o_ready_assignment : for k in 0 to 5 loop
        rmq_snk_o(0)(k).ready <= not miso_fifo_wr_full(k);
        rmq_snk_o(0)(k).pkt_ready <= '1';
      end loop;

      rmq_snk_o(1)(0).ready <= not miso_fifo_wr_full(6);
      rmq_snk_o(1)(0).pkt_ready <= '1';

      --! Write enable of the MISO FIFOs
      miso_fifo_write_enable : for k in 0 to 6 loop
        miso_wr_en(k) <= miso_fifo_data_in(k)(34) and (not miso_fifo_wr_full(k));
        if miso_wr_en(k) = '1' then
          s_rmq_id <= k;
        end if;
      end loop;

      --! Read enable of the MISO FIFO
      miso_fifo_read_enable : for k in 0 to 6 loop
        miso_rd_en(k) <= rd_en and (not miso_fifo_rd_empty(k));
      end loop;

   end process;

  rd_en <= '1' when (s_rd_state = RD_HEADER or s_rd_state = LOAD_HWORD or s_rd_state = LOAD_LWORD) else '0';

  spi_miso_data    <= miso_data_reg when s_ready = '1';
  spi_data_valid_i <= s_data_valid;

  -- Flag that indicates if the operation will be write or read to/from RMQs
  s_oper_en <= s_operation when (s_rd_state = IDLE and s_wr_state = IDLE) else '0' when s_rd_state = END_OF_FRAME;

--  s_rmq_id  <= to_integer(unsigned(wr_rmq_id));

  --! Capture the positive edge of the spi-ready
  cmp_ready_o_edge_detect : entity work.gc_edge_detect
    generic map(
      g_ASYNC_RST  => false,
      g_PULSE_EDGE => "positive",
      g_CLOCK_EDGE => "positive")
    port map(
      clk_i        => clk_i,
      rst_n_i      => rst_n_i,
      data_i       => spi_ready,
      pulse_o      => s_ready);

end architecture rtl;